﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Web.Script.Serialization;
using System.Windows.Forms;

namespace VrpSolverClean
{
    class MatematickaVzdalenost
    {
        const double PIx = 3.141592653589793;
        const double RADIUS = 6371.0088;

        /// <summary>
        /// Convert degrees to Radians
        /// </summary>
        /// <param name="x">Degrees</param>
        /// <returns>The equivalent in radians</returns>
        public static double Radians(double x)
        {
            return x * PIx / 180;
        }

        /// <summary>
        /// Calculate the distance between two places.
        /// </summary>
        /// <param name="lon1"></param>
        /// <param name="lat1"></param>
        /// <param name="lon2"></param>
        /// <param name="lat2"></param>
        /// <returns></returns>
        public static double DistanceBetweenPlaces(double lon1, double lat1, double lon2, double lat2)
        {
            double dlon = Radians(lon2 - lon1);
            double dlat = Radians(lat2 - lat1);

            double a = (Math.Sin(dlat / 2) * Math.Sin(dlat / 2)) + Math.Cos(Radians(lat1)) * Math.Cos(Radians(lat2)) * (Math.Sin(dlon / 2) * Math.Sin(dlon / 2));
            double angle = 2 * Math.Atan2(Math.Sqrt(a), Math.Sqrt(1 - a));
            return angle * RADIUS;
        }
    }

    class NejblizsiSoused
    {
        public static Tuple<List<int>, double> NejlepsiCestaMeziZakazniky = NajdiNejlepsiCestuMeziExistujicimiZakazniky();

        private static List<Tuple<Zakaznik,Zakaznik,double>> najdiVzdalenostiMeziExistujicimiZakazniky()
        {
            // proměnná obsahující list kombinací všech zákazníků a vzdáleností mezi nimi. Toto bude použito jako základ pro výpočet trasy mezi nimi.
            var vzdalenosti = (from zakaznik1 in Databaze.Zakaznici
                               from zakaznik2 in Databaze.Zakaznici
                               where zakaznik1.ID < zakaznik2.ID
                               select new Tuple<Zakaznik, Zakaznik, double>(
                                   zakaznik1,
                                   zakaznik2,
                                   Google.SpocitejVzdalenost(zakaznik1.GeoLat, zakaznik1.GeoLon, zakaznik2.GeoLat, zakaznik2.GeoLon)))
                                .ToList();

            return vzdalenosti;
        }

        public static Tuple<List<int>, double> NajdiNejlepsiCestuMeziExistujicimiZakazniky()
        {
            //zacatek algoritmu pro vypocet trasy pomoci nejblizsiho souseda
            var delkaCesty = 0.0; //promenna obsahujici delku finalni trasy
            var delkaKSousedovi = 100000.0; //cena cesty k dalsimu zakaznikovi
            var stavajiciMisto = 0; //stavajici zakaznik
            var soused = 0; //pristi zakaznik
            var navstiveno = new List<int>(); //seznam ID jiz navstivenych zakazniku, dulezite pro zjisteni, kdo byl jiz navstiven a kam je stale potreba jet
            var vzdalenosti = najdiVzdalenostiMeziExistujicimiZakazniky();

            while (navstiveno.Count() < Databaze.Zakaznici.Count - 1)
            {
                //pro kazdy spoj ze seznamu vsech spoju 
                foreach (var spoj in vzdalenosti)
                {
                    //pokud je stavajici zakaznik na prvnim miste v listu spoju, nebyl jeste navstiven a cena spoje je nizsi nez cena k sousedovi
                    if (spoj.Item1.ID == stavajiciMisto &&
                        !navstiveno.Contains(spoj.Item2.ID) &&
                        delkaKSousedovi > spoj.Item3)
                    {
                        soused = spoj.Item2.ID;
                        delkaKSousedovi = spoj.Item3;
                    }
                    else if (spoj.Item2.ID == stavajiciMisto &&
                        !navstiveno.Contains(spoj.Item1.ID) &&
                        delkaKSousedovi > spoj.Item3)
                    {
                        soused = spoj.Item1.ID;
                        delkaKSousedovi = spoj.Item3;
                    }
                }

                //po nalezeni nejblizsiho souseda je cesta prodlouzena o cestu k sousedovi, soused je prirazen jako stavajici misto a delka k sousedovi je vynulovana
                navstiveno.Add(stavajiciMisto);
                stavajiciMisto = soused;
                delkaCesty += delkaKSousedovi;
                delkaKSousedovi = 100000.0;
                if (navstiveno.Count() == Databaze.Zakaznici.Count - 1)
                {
                    navstiveno.Add(soused);
                }
            }

            var kompletniCesta = new Tuple<List<int>, double>(navstiveno, delkaCesty);

            return kompletniCesta;
        }
    }

    class Google
    {
        // API key od Google, pouze pro tento projekt, pokud by se program distribuoval, musel by byt zmenen, aby musel byt klic zadan pri startu programu.
        private static string googleApiKey = "AIzaSyBNS1rrvkgpMTf2Hh8JHGam6r2yFeCiXHM";
        // outputFormat muze byt json nebo xml
        private static string outputFormat = "json";
        private static string scheme = "https";
        private static string host = "maps.googleapis.com";


        /// <summary>
        /// Vrátí string s URL ke k obrázku mapy, na které je vykreslena ideálni trasa mezi zákazníky.
        /// </summary>
        /// <returns></returns>
        public static Uri GetMapUri()
        {
            var mapSize = "640x480";
            var mapType = "roadmap"; // roadmap, satellite, hybrid, and terrain

            var markers = "markers=color:green";

            string path = "path=color:blue";

            foreach (var cesta in NejblizsiSoused.NejlepsiCestaMeziZakazniky.Item1)
            {
                foreach (var zakaznik in Databaze.Zakaznici)
                {
                    if (zakaznik.ID == cesta)
                    {
                        path += $"|{zakaznik.GeoLat},{zakaznik.GeoLon}";
                        markers += $"|{zakaznik.GeoLat},{zakaznik.GeoLon}";
                        break;
                    }
                }
            }

            var builder = new UriBuilder
            {
                Scheme = scheme,
                Host = host,
                Path = "/maps/api/staticmap",
                Query = $"size={mapSize}&maptype={mapType}&{markers}&{path}&key={googleApiKey}"
            };

            var uri = builder.Uri;
            return uri;
        }

        /// <summary>
        /// Funkce potřebuje souřadnice dvou míst. Připojí se ke Google API pro výpočet vzdálenosti a zjistí vzdálenost mezi zadanými místy.
        /// </summary>
        /// <param name="sLat"></param>
        /// <param name="sLon"></param>
        /// <param name="dLat"></param>
        /// <param name="dLon"></param>
        /// <returns>
        /// Vrátí vzdálenost mezi dvěma místy ze zadání (v kilometrech).
        /// </returns>
        public static double SpocitejVzdalenost(double sLat, double sLon, double dLat, double dLon)
        {
            UriBuilder builder = new UriBuilder
            {
                Scheme = scheme,
                Host = host,
                Path = $"/maps/api/distancematrix/{outputFormat}",
                Query = $"origins={sLat},{sLon}&destinations={dLat},{dLon}&key={googleApiKey}"
            };
            var uri = builder.Uri;

            string json;
            try
            {
                json = new WebClient().DownloadString(uri);
            }
            catch (Exception)
            {
                MessageBox.Show("Neni pripojeni k internetu, pokud nejste pripojeni k internetu, pouzijte jinou moznost vypoctu vzdalenosti");
                throw;
            }

            var serializer = new JavaScriptSerializer();
            dynamic result = serializer.DeserializeObject(json);
            string vzdalenost = result["rows"][0]["elements"][0]["distance"]["text"];

            return Convert.ToDouble(vzdalenost.TrimEnd(" km".ToCharArray()));
        }

        /// <summary>
        /// Po zadání adresy složené z ulice, čísla popisného a města - vrátí zeměpisné souřadnice jako Dictionary s klíčovou hodnotou string a hodnotou double.
        /// </summary>
        /// <param name="ulice"></param>
        /// <param name="cp"></param>
        /// <param name="mesto"></param>
        /// <returns>
        /// Vrátí souřadnice zadaného místa podle adresy.
        /// </returns>
        public static Dictionary<string, double> ZjistiSouradnice(string ulice, string cp, string mesto)
        {
            var builder = new UriBuilder
            {
                Scheme = scheme,
                Host = host,
                Path = $"/maps/api/geocode/{outputFormat}",
                Query = $"address={cp}+{ulice.Replace(" ", "+")}+,{mesto.Replace(" ", "+")}&key={googleApiKey}"
            };

            var uri = builder.Uri;
            
            string json;
            try
            {
                json = new WebClient().DownloadString(uri);
            }
            catch (Exception)
            {
                MessageBox.Show("Není připojení k internetu, pokud nejste připojeni k internetu, použijte jinou možnost výpočtu vzdálenosti");
                throw;
            }

            var serializer = new JavaScriptSerializer();
            dynamic result = serializer.DeserializeObject(json);
            double geoLat = (double) result["results"][0]["geometry"]["location"]["lat"];
            double geoLon = (double) result["results"][0]["geometry"]["location"]["lng"];

            var souradnice = new Dictionary<string, double>
            {
                { "Lat", geoLat },
                { "Lon", geoLon }
            };

            return souradnice;
        }
    }
}
